@@ -20,6 +20,7 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include "config.h"
 #include <stdlib.h>
 #include <string.h>
 #include <strings.h>
@@ -29,12 +30,16 @@
 #include <gdk/gdkx.h>
 #include <gtk/gtk.h>
 #include <X11/Xatom.h>
+#ifdef HAVE_XFIXES
+#include <X11/extensions/Xfixes.h>
+#endif
 #include "magnifier.h"
 #include "magnifier-private.h"
 #include "zoom-region.h"
 #include "zoom-region-private.h"
 #include "GNOME_Magnifier.h"
 
+
 #define MAG_WINDOW_OVERRIDE_REDIRECT
 
 static GObjectClass *parent_class = NULL;
@@ -56,6 +61,11 @@ enum {
 } PropIdx;
 
 static int _x_error = 0;
+static int fixes_event_base = 0, fixes_error_base;
+static Display *cursor_client_connection;
+
+static void magnifier_transform_cursor (Magnifier *magnifier);
+static void magnifier_init_cursor_set (Magnifier *magnifier, gchar *cursor_set);
 
 gboolean
 magnifier_error_check (void)
@@ -80,6 +90,234 @@ magnifier_x_error_handler (Display	 *display,
 	return 0;
 }
 
+static void
+magnifier_zoom_regions_mark_dirty (Magnifier *magnifier, GNOME_Magnifier_RectBounds rect_bounds)
+{
+	GList *list;
+
+	g_assert (magnifier);
+	list = magnifier->zoom_regions;
+	while (list) 
+	{
+		/* propagate the expose events to the zoom regions */
+		GNOME_Magnifier_ZoomRegion zoom_region;
+		CORBA_Environment ev;
+		zoom_region = list->data;
+		CORBA_exception_init (&ev);
+		if (zoom_region)
+			GNOME_Magnifier_ZoomRegion_markDirty (CORBA_Object_duplicate (zoom_region, &ev),
+							      &rect_bounds,
+							      &ev);
+		list = g_list_next (list);
+	}
+}
+
+void
+magnifier_set_cursor_from_pixbuf (Magnifier *magnifier, GdkPixbuf *cursor_pixbuf)
+{
+	GdkPixmap *pixmap, *mask;
+	gint width, height;
+	GdkGC *gc;
+	GdkDrawable *drawable = magnifier->priv->w->window;
+
+	if (magnifier->priv->cursor) {
+		g_object_unref (magnifier->priv->cursor);
+		magnifier->priv->cursor = NULL;
+	}
+	if (drawable && cursor_pixbuf)
+	{
+		const gchar *xhot_string = NULL, *yhot_string = NULL;
+		width = gdk_pixbuf_get_width (cursor_pixbuf);
+		height = gdk_pixbuf_get_height (cursor_pixbuf);
+		pixmap = gdk_pixmap_new (drawable, width, height, -1);
+		gc = gdk_gc_new (pixmap);
+		gdk_draw_pixbuf (pixmap, gc, cursor_pixbuf, 0, 0, 0, 0, 
+				 width, height,
+				 GDK_RGB_DITHER_NONE, 0, 0);
+		mask = gdk_pixmap_new (drawable, width, height, 1);
+		gdk_pixbuf_render_threshold_alpha (cursor_pixbuf, mask, 0, 0, 0, 0, 
+						   width, height,
+						   200);
+		gdk_gc_unref (gc);
+		magnifier->priv->cursor = pixmap;
+		magnifier->priv->cursor_mask = mask;
+		xhot_string = gdk_pixbuf_get_option (cursor_pixbuf,"x_hot");
+		yhot_string = gdk_pixbuf_get_option (cursor_pixbuf,"y_hot");
+		if (xhot_string) magnifier->cursor_hotspot.x = atoi (xhot_string);
+		if (yhot_string) magnifier->cursor_hotspot.y = atoi (yhot_string);
+
+		if (pixmap) {
+			gdk_drawable_get_size (pixmap,
+					       &magnifier->priv->cursor_default_size_x,
+					       &magnifier->priv->cursor_default_size_y);
+			magnifier->priv->cursor_hotspot_x = magnifier->cursor_hotspot.x;
+			magnifier->priv->cursor_hotspot_y = magnifier->cursor_hotspot.y;
+		}
+	}
+}
+
+
+void
+magnifier_free_cursor_pixels (guchar *pixels, gpointer data)
+{
+    /* XFree (data); FIXME why doesn't this work properly? */
+}
+
+GdkPixbuf *
+magnifier_get_source_pixbuf (Magnifier *magnifier)
+{
+#ifdef HAVE_XFIXES
+	XFixesCursorImage *cursor_image = XFixesGetCursorImage (cursor_client_connection);
+        GdkPixbuf *cursor_pixbuf = NULL;
+	gchar s[6];
+	if (cursor_image)
+	{
+		cursor_pixbuf = gdk_pixbuf_new_from_data ((guchar *) cursor_image->pixels, 
+							  GDK_COLORSPACE_RGB, 
+							  TRUE, 8,  
+							  cursor_image->width, cursor_image->height,
+							  cursor_image->width * 4, 
+							  magnifier_free_cursor_pixels, 
+							  cursor_image );
+		gdk_pixbuf_set_option (cursor_pixbuf, "x_hot", 
+				       g_ascii_dtostr (s, 6, (gdouble) cursor_image->xhot));
+		gdk_pixbuf_set_option (cursor_pixbuf, "y_hot", 
+				       g_ascii_dtostr (s, 6, (gdouble) cursor_image->yhot));
+	}
+	return cursor_pixbuf;
+#else
+	return NULL;
+#endif
+}
+
+GdkPixbuf *
+magnifier_get_pixbuf_for_name (Magnifier *magnifier, const gchar *cursor_name)
+{
+    GdkPixbuf *retval = NULL;
+    g_message ("cursor %s", cursor_name);
+    if (magnifier->priv->cursorlist) 
+	    retval = g_hash_table_lookup (magnifier->priv->cursorlist, cursor_name);
+    if (retval) 
+	    g_object_ref (retval);
+    return retval;
+}
+
+void
+magnifier_set_cursor_pixmap_by_name (Magnifier *magnifier, const gchar *cursor_name, 
+				     gboolean source_fallback)
+{
+	GdkPixbuf *pixbuf;
+	/* search local table; if not found, use source screen's cursor if source_fallback is TRUE */
+	if ((pixbuf = magnifier_get_pixbuf_for_name (magnifier, cursor_name)) == NULL) {
+#ifndef HAVE_XFIXES
+		source_fallback = FALSE;
+#endif
+		if (source_fallback == TRUE)
+		{
+			pixbuf = magnifier_get_source_pixbuf (magnifier);
+		}
+		else
+		{
+			pixbuf = magnifier_get_pixbuf_for_name (magnifier, "default");
+		}
+	}
+	magnifier_set_cursor_from_pixbuf (magnifier, pixbuf);
+	if (pixbuf) gdk_pixbuf_unref (pixbuf);
+}
+gboolean
+magnifier_cursor_notify (GIOChannel *source, GIOCondition condition, gpointer data)
+{
+#ifdef HAVE_XFIXES
+  XEvent ev;
+  Magnifier *magnifier = (Magnifier *) data;
+  XFixesCursorNotifyEvent  *cev;
+
+  /* TODO: don't ask for name unless xfixes version >= 2 available */
+  do
+  {
+      XNextEvent(cursor_client_connection, &ev);
+      if (ev.type == fixes_event_base + XFixesCursorNotify) 
+      {
+	  cev = (XFixesCursorNotifyEvent *) &ev;
+      }
+  } while (XPending (cursor_client_connection));
+
+  if (magnifier->priv->use_source_cursor) 
+  {
+      GdkPixbuf *cursor_pixbuf = magnifier_get_source_pixbuf (magnifier);
+      magnifier_set_cursor_from_pixbuf (magnifier, cursor_pixbuf);
+      if (cursor_pixbuf) gdk_pixbuf_unref (cursor_pixbuf);
+  }
+  else
+  {
+      magnifier_set_cursor_pixmap_by_name (magnifier, 
+					   gdk_x11_get_xatom_name (cev->cursor_name),
+					   TRUE);
+  }
+
+  magnifier_transform_cursor (magnifier);
+#ifdef DEBUG_CURSOR
+  g_message ("cursor changed: subtype=%d, cursor_serial=%lu, name=[%x] %s\n",
+	     (int) cev->subtype, cev->cursor_serial, (int) cev->cursor_name,
+	     gdk_x11_get_xatom_name (cev->cursor_name));
+#endif
+
+  return TRUE;
+#else
+  return FALSE;
+#endif
+}
+
+gboolean
+magnifier_cursor_notification_init (Magnifier *magnifier)
+{
+#ifdef HAVE_XFIXES
+    GIOChannel *ioc;
+    int fd;
+    Display *dpy;
+    Window rootwin;
+
+    dpy = GDK_DISPLAY_XDISPLAY (magnifier->source_display);
+    cursor_client_connection = XOpenDisplay (0);
+    rootwin = GDK_WINDOW_XWINDOW (magnifier->priv->root);
+
+    if (!XFixesQueryExtension (cursor_client_connection, &fixes_event_base, &fixes_error_base))
+    {
+	g_warning ("XFixes extension not currently active.\n");
+	return FALSE;
+    }
+    else
+    {
+	XFixesSelectCursorInput (cursor_client_connection, rootwin, XFixesDisplayCursorNotifyMask);
+	fd = ConnectionNumber (cursor_client_connection);
+	ioc = g_io_channel_unix_new (fd);
+	g_io_add_watch (ioc, G_IO_IN | G_IO_HUP | G_IO_PRI | G_IO_ERR, magnifier_cursor_notify, 
+			magnifier);
+	g_io_channel_unref (ioc); 
+	g_message ("added event source to xfixes cursor-notify connection");
+ 	XFlush (cursor_client_connection); 
+   }
+    return TRUE;
+#else
+    g_warning ("this copy of gnome-mag was built without xfixes extension support.\n");
+    return FALSE;
+#endif
+}
+
+void
+magnifier_notify_damage (Magnifier *magnifier, XRectangle *rect)
+{
+	GNOME_Magnifier_RectBounds rect_bounds;
+	rect_bounds.x1 = rect->x;
+	rect_bounds.y1 = rect->y;
+	rect_bounds.x2 = rect->x + rect->width;
+	rect_bounds.y2 = rect->y + rect->height;
+#ifdef DEBUG_DAMAGE
+	g_message ("damage");
+#endif
+	magnifier_zoom_regions_mark_dirty (magnifier, rect_bounds);
+}
+
 GdkFilterReturn
 magnifier_expose_filter (GdkXEvent *xevent, GdkEvent *event, gpointer data)
 {
@@ -95,32 +333,21 @@ magnifier_expose_filter (GdkXEvent *xevent, GdkEvent *event, gpointer data)
 		rect_bounds.y1 = rect.y;
 		rect_bounds.x2 = rect.x + rect.width;
 		rect_bounds.y2 = rect.y + rect.height;
-		
-		g_assert (magnifier);
-		list = magnifier->zoom_regions;
-		
-		while (list) 
-		{
-			/* propagate the expose events to the zoom regions */
-			ZoomRegion *zoom_region;
-			CORBA_Environment ev;
-			zoom_region = list->data;
-			CORBA_exception_init (&ev);
-			GNOME_Magnifier_ZoomRegion_markDirty (BONOBO_OBJREF (zoom_region),
-							      &rect_bounds,
-							      &ev);
-			list = g_list_next (list);
-		}
+		magnifier_zoom_regions_mark_dirty (magnifier, rect_bounds);
+
 		return GDK_FILTER_TRANSLATE;
 	}
 	return GDK_FILTER_CONTINUE;
 }
 
 static void
-magnifier_set_expose_listener (Magnifier *magnifier, GdkWindow *root)
+magnifier_set_extension_listeners (Magnifier *magnifier, GdkWindow *root)
 {
-	/* todo: remove listener from 'old' root if one was already present */
+	g_message ("adding expose listener on root");
+	/* TODO: remove listener from 'old' root if one was already present */	
 	gdk_window_add_filter (root, magnifier_expose_filter, NULL);
+	magnifier_damage_client_init (magnifier);
+	magnifier_cursor_notification_init (magnifier);
 }
 
 static void
@@ -134,7 +361,7 @@ magnifier_realize (GtkWidget *widget)
 	
 	static gboolean initialized = FALSE;
 
-#ifdef MAG_WINDOW_OVERRIDE_REDIRECT	
+#ifndef MAG_WINDOW_OVERRIDE_REDIRECT	
 	if (!initialized) {
 		wm_window_protocols[0] = gdk_x11_get_xatom_by_name_for_display (target_display,
 										"WM_DELETE_WINDOW");
@@ -171,7 +398,6 @@ magnifier_realize (GtkWidget *widget)
 	 */
 
 	XSetErrorHandler (magnifier_x_error_handler);
-
 }
 
 GdkWindow*
@@ -185,13 +411,6 @@ magnifier_get_root (Magnifier *magnifier)
 	return magnifier->priv->root;
 }
 
-static void
-magnifier_exit (GtkObject *object)
-{
-	gtk_main_quit ();
-	exit (0);
-}
-
 static gint
 magnifier_parse_display_name (Magnifier *magnifier, gchar *full_display_string,
 			      gchar **display_name)
@@ -213,6 +432,142 @@ magnifier_parse_display_name (Magnifier *magnifier, gchar *full_display_string,
 	return 0;
 }
 
+static void
+magnifier_get_display_rect_bounds (Magnifier *magnifier, GNOME_Magnifier_RectBounds *rect_bounds, gboolean is_target)
+{
+    if (is_target)
+    {
+	rect_bounds->x1 = 0;
+	rect_bounds->x2 = gdk_screen_get_width (
+	    gdk_display_get_screen (magnifier->target_display,
+				    magnifier->target_screen_num));
+	rect_bounds->y1 = 0;
+	rect_bounds->y2 = gdk_screen_get_height (
+	    gdk_display_get_screen (magnifier->target_display,
+				    magnifier->target_screen_num));
+
+    }
+    else 
+    {
+	rect_bounds->x1 = 0;
+	rect_bounds->x2 = gdk_screen_get_width (
+	    gdk_display_get_screen (magnifier->source_display,
+				    magnifier->source_screen_num));
+	rect_bounds->y1 = 0;
+	rect_bounds->y2 = gdk_screen_get_height (
+	    gdk_display_get_screen (magnifier->source_display,
+				    magnifier->source_screen_num));
+
+    }
+}
+
+static void
+magnifier_unref_zoom_region (gpointer data, gpointer user_data)
+{
+/*	Magnifier *magnifier = user_data; NOT USED */
+	CORBA_Environment ev;
+	GNOME_Magnifier_ZoomRegion zoom_region = data;
+	CORBA_exception_init (&ev);
+	GNOME_Magnifier_ZoomRegion_dispose (zoom_region, &ev);
+}
+
+static void
+magnifier_reparent_zoom_regions (Magnifier *magnifier)
+{
+    GList *list;
+    
+    g_assert (magnifier);
+    list = magnifier->zoom_regions;
+    while (list) 
+    {
+	GNOME_Magnifier_ZoomRegion zoom_region;
+	CORBA_Environment ev;
+	zoom_region = list->data;
+	CORBA_exception_init (&ev);
+	if (zoom_region)
+	{
+	    GNOME_Magnifier_ZoomRegion new_region;
+	    Bonobo_PropertyBag properties, new_properties;
+	    GNOME_Magnifier_RectBounds rectbounds, viewport;
+	    CORBA_any *value;
+	    rectbounds = GNOME_Magnifier_ZoomRegion_getROI (zoom_region, &ev);
+	    properties = GNOME_Magnifier_ZoomRegion_getProperties (zoom_region, &ev);
+	    value = bonobo_pbclient_get_value (properties, "viewport", TC_GNOME_Magnifier_RectBounds, &ev);
+	    memcpy (&viewport, value->_value, sizeof (GNOME_Magnifier_RectBounds));
+	    fprintf (stderr, "reparenting zoomer with viewport %d,%d - %d,%d\n", viewport.x1, viewport.y1, viewport.x2, viewport.y2);
+	    CORBA_free (value);
+	    new_region = GNOME_Magnifier_Magnifier_createZoomRegion (BONOBO_OBJREF (magnifier), 0, 0, &rectbounds, &viewport, &ev);
+	    new_properties = GNOME_Magnifier_ZoomRegion_getProperties (new_region, &ev);
+	    bonobo_pbclient_set_boolean (new_properties, "is-managed", 
+					 bonobo_pbclient_get_boolean (properties, "is-managed", NULL), NULL);
+	    bonobo_pbclient_set_long (new_properties, "smooth-scroll-policy", 
+				       bonobo_pbclient_get_long (properties, "smooth-scroll-policy", NULL), NULL);
+	    bonobo_pbclient_set_float (new_properties, "contrast", 
+					 bonobo_pbclient_get_float (properties, "contrast", NULL), NULL);
+	    bonobo_pbclient_set_float (new_properties, "mag-factor-x", 
+					 bonobo_pbclient_get_float (properties, "mag-factor-x", NULL), NULL);
+	    bonobo_pbclient_set_float (new_properties, "mag-factor-y", 
+					 bonobo_pbclient_get_float (properties, "mag-factor-y", NULL), NULL);
+	    bonobo_pbclient_set_long (new_properties, "x-alignment", 
+					 bonobo_pbclient_get_long (properties, "x-alignment", NULL), NULL);
+	    bonobo_pbclient_set_long (new_properties, "y-alignment", 
+					 bonobo_pbclient_get_long (properties, "y-alignment", NULL), NULL);
+	    bonobo_pbclient_set_long (new_properties, "border-color", 
+				      bonobo_pbclient_get_long (properties, "border-color", NULL), NULL); 
+	    bonobo_pbclient_set_long (new_properties, "border-size", 
+				      bonobo_pbclient_get_long (properties, "border-size", NULL), NULL);
+	    bonobo_pbclient_set_string (new_properties, "smoothing-type", 
+					bonobo_pbclient_get_string (properties, "smoothing-type", NULL), NULL); 
+	    bonobo_pbclient_set_boolean (new_properties, "inverse-video", 
+					 bonobo_pbclient_get_boolean (properties, "inverse-video", NULL), NULL); 
+	    bonobo_object_release_unref (properties, &ev);
+	    bonobo_object_release_unref (new_properties, &ev);
+	    list->data = new_region;
+	    magnifier_unref_zoom_region ((gpointer) zoom_region, NULL);
+	}
+	list = g_list_next (list);
+    }   
+}
+
+static void
+magnifier_init_display (Magnifier *magnifier, gchar *display_name, gboolean is_target)
+{
+    if (is_target)
+    {
+	magnifier->target_screen_num =
+	    magnifier_parse_display_name (magnifier,
+					  display_name,
+					  NULL);
+	magnifier->target_display =
+	    gdk_display_open (display_name);
+	magnifier->priv->root =
+	    gdk_screen_get_root_window (
+		gdk_display_get_screen (
+		    magnifier->target_display,
+		    magnifier->target_screen_num));
+    }
+    else 
+    {
+	magnifier->source_screen_num =
+	    magnifier_parse_display_name (magnifier,
+					  display_name,
+					  NULL);
+	magnifier->source_display =
+	    gdk_display_open (display_name);
+	magnifier->priv->root =
+	    gdk_screen_get_root_window (
+		gdk_display_get_screen (
+		    magnifier->source_display,
+		    magnifier->source_screen_num));
+    }
+}
+
+static void
+magnifier_exit (GtkObject *object)
+{
+	gtk_main_quit ();
+	exit (0);
+}
 
 #define GET_PIXEL(a,i,j,s,b) \
 (*(guint32 *)(memcpy (b,(a) + ((j) * s + (i) * pixel_size_t), pixel_size_t)))
@@ -237,7 +592,7 @@ magnifier_recolor_pixbuf (Magnifier *magnifier, GdkPixbuf *pixbuf)
 	for (j = 0; j < h; ++j) {
 		for (i = 0; i < w; ++i) {
 			pixval = GET_PIXEL (pixels, i, j, rowstride, &pixval);
-			if ((pixval & 0xFFFFFF) == 0)
+			if ((pixval & 0x808080) == 0)
 			{
 				pixval = cursor_color;
 				PUT_PIXEL (pixels, i, j, rowstride,
@@ -253,7 +608,7 @@ magnifier_transform_cursor (Magnifier *magnifier)
 	if (magnifier->priv->cursor) /* don't do this if cursor isn't intialized yet */
 	{
 		int width, height;
-		int size;
+		int size_x, size_y;
 		GdkPixbuf *scaled_cursor_pixbuf;
 		GdkPixbuf *scaled_mask_pixbuf;
 		GdkPixbuf *cursor_pixbuf;
@@ -263,13 +618,19 @@ magnifier_transform_cursor (Magnifier *magnifier)
 		GdkGC *cgc;
 		GdkGC *mgc;
 
-		if (magnifier->cursor_size)
-			size = magnifier->cursor_size;
+		if (magnifier->cursor_size_x)
+		{
+			size_x = magnifier->cursor_size_x;
+		}
 		else
-			size = magnifier->priv->cursor_default_size * 
+		{
+			size_x = magnifier->priv->cursor_default_size_x * 
 			       magnifier->cursor_scale_factor;
+			size_y = magnifier->priv->cursor_default_size_y * 
+			       magnifier->cursor_scale_factor;
+		}
 		gdk_drawable_get_size (magnifier->priv->cursor, &width, &height);
-		if ((size == width) 
+		if ((size_x == width) && (size_y == height) 
 		    && (magnifier->cursor_color == 0xFF000000))
 			return; /* nothing changes */
 
@@ -285,38 +646,38 @@ magnifier_transform_cursor (Magnifier *magnifier)
 							    NULL, 0, 0, 0, 0,
 							    width, height);
 		scaled_cursor_pixbuf = gdk_pixbuf_scale_simple (
-			cursor_pixbuf, size, size, GDK_INTERP_NEAREST);
+			cursor_pixbuf, size_x, size_y, GDK_INTERP_NEAREST);
 		
-		magnifier->cursor_hotspot.x = magnifier->priv->cursor_hotspot_x * size 
-					    / magnifier->priv->cursor_default_size;
-		magnifier->cursor_hotspot.y = magnifier->priv->cursor_hotspot_y * size 
-					    / magnifier->priv->cursor_default_size;
+		magnifier->cursor_hotspot.x = magnifier->priv->cursor_hotspot_x * size_x 
+					    / magnifier->priv->cursor_default_size_x;
+		magnifier->cursor_hotspot.y = magnifier->priv->cursor_hotspot_y * size_y 
+					    / magnifier->priv->cursor_default_size_y;
 					    
 		scaled_mask_pixbuf = gdk_pixbuf_scale_simple (
-			mask_pixbuf, size, size, GDK_INTERP_NEAREST);
+			mask_pixbuf, size_x, size_y, GDK_INTERP_NEAREST);
 		g_object_unref (cursor_pixbuf);
 		g_object_unref (mask_pixbuf);
 		g_object_unref (cursor_pixmap);
 		g_object_unref (mask_pixmap);
 		magnifier->priv->cursor = gdk_pixmap_new (
 			magnifier->priv->w->window,
-			size, size,
+			size_x, size_y,
 			-1);
 		magnifier->priv->cursor_mask = gdk_pixmap_new (
 			magnifier->priv->w->window,
-			size, size,
+			size_x, size_y,
 			1);
 
 		gdk_draw_pixbuf (magnifier->priv->cursor,
 				 cgc,
 				 scaled_cursor_pixbuf,
-				 0, 0, 0, 0, size, size,
+				 0, 0, 0, 0, size_x, size_y,
 				 GDK_RGB_DITHER_NONE, 0, 0 );
 		scaled_mask_pixbuf = gdk_pixbuf_add_alpha (
 			scaled_mask_pixbuf, True, 0, 0, 0);
 		gdk_pixbuf_render_threshold_alpha (scaled_mask_pixbuf,
 						   magnifier->priv->cursor_mask,
-						   0, 0, 0, 0, size, size,
+						   0, 0, 0, 0, size_x, size_y,
 						   0x80);
 		g_object_unref (scaled_cursor_pixbuf);
 		g_object_unref (scaled_mask_pixbuf);
@@ -326,73 +687,70 @@ magnifier_transform_cursor (Magnifier *magnifier)
 }
 
 static void
-magnifier_unref_zoom_region (gpointer data, gpointer user_data)
-{
-/*	Magnifier *magnifier = user_data; NOT USED */
-	CORBA_Environment ev;
-	GNOME_Magnifier_ZoomRegion zoom_region = data;
-	CORBA_exception_init (&ev);
-	GNOME_Magnifier_ZoomRegion_dispose (zoom_region, &ev);
-}
-
-static void
-magnifier_init_cursor (Magnifier *magnifier)
+magnifier_init_cursor_set (Magnifier *magnifier, gchar *cursor_set)
 {
 	GdkPixbuf	*cursor_pixbuf = NULL;
 	/*
 	 * we check the cursor-set property string here,
 	 * and create/apply the appropriate cursor settings
 	 */
-	    
-	if (!strcmp (magnifier->cursor_set, "none"))
+	magnifier->cursor_set = cursor_set;
+#ifdef HAVE_XFIXES	
+	magnifier->priv->use_source_cursor = 
+	    (!strcmp (cursor_set, "default") && 
+	     (fixes_event_base != 0));
+#else
+	magnifier->priv->use_source_cursor = FALSE;
+#endif
+	if (magnifier->priv->use_source_cursor) return;
+
+	if (!strcmp (magnifier->cursor_set, "none")) {
 		magnifier->priv->cursor = NULL;
+		return;
+	}
 	else 
 	{
-		gchar *xpm_data_file = g_strconcat (CURSORSDIR,"/",
-						    magnifier->cursor_set,
-						    "-cursor.xpm", NULL);
+		GDir *cursor_dir;
+		const gchar *filename;
+		gchar *cursor_dirname;
 
-		cursor_pixbuf = gdk_pixbuf_new_from_file (xpm_data_file, NULL);
-		
-		if (cursor_pixbuf)
+		if (magnifier->priv->cursorlist)
 		{
-		    magnifier->cursor_hotspot.x = atoi (gdk_pixbuf_get_option (cursor_pixbuf,"x_hot") );
-		    magnifier->cursor_hotspot.y = atoi (gdk_pixbuf_get_option (cursor_pixbuf,"y_hot") );
-
-		    fprintf (stderr,"\nHOTSPOT  %d %d\n",
-				magnifier->cursor_hotspot.x,
-				magnifier->cursor_hotspot.y);
-		    g_object_unref (G_OBJECT (cursor_pixbuf));
+			g_hash_table_destroy (magnifier->priv->cursorlist);
 		}
-		else
-		    fprintf (stderr,"\nHOTSPOT failed\n");
+		magnifier->priv->cursorlist = g_hash_table_new_full (g_str_hash, g_str_equal,
+								     g_free, g_object_unref);
 
-		if (magnifier->priv->cursor) {
-			g_object_unref (magnifier->priv->cursor);
-			magnifier->priv->cursor = NULL;
-		}
-		if (magnifier->priv->w->window)
+		cursor_dirname = g_strconcat (CURSORSDIR, "/", magnifier->cursor_set, NULL);
+		cursor_dir = g_dir_open (cursor_dirname, 0, NULL);
+		/* assignment, not comparison, is intentional */
+		g_message ("cursor_dir=%x", cursor_dir);
+		while (cursor_dir && (filename = g_dir_read_name (cursor_dir)) != NULL) 
 		{
-			magnifier->priv->cursor = 
-				gdk_pixmap_create_from_xpm (
-					magnifier->priv->w->window,
-					&magnifier->priv->cursor_mask,
-					NULL,
-					xpm_data_file);
-
-			if (magnifier->priv->cursor) {
-				gdk_drawable_get_size (magnifier->priv->cursor,
-						       &magnifier->priv->cursor_default_size,
-						       &magnifier->priv->cursor_default_size);
-				magnifier->priv->cursor_hotspot_x = magnifier->cursor_hotspot.x;
-				magnifier->priv->cursor_hotspot_y = magnifier->cursor_hotspot.y;
-				fprintf (stderr, "getting cursor %p, mask %p\n",
-					 magnifier->priv->cursor, magnifier->priv->cursor_mask);
+			if (filename) 
+			{
+				gchar *path = g_strconcat (cursor_dirname, "/", filename, NULL);
+				GdkPixbuf *pixbuf = gdk_pixbuf_new_from_file (path, NULL);
+				if (pixbuf)
+				{
+					/* add this pixbuf and its name to our list */
+					gchar **sv, *cname;
+					cname = g_path_get_basename (filename);
+					sv = g_strsplit (cname, ".", 2);
+					g_hash_table_insert (magnifier->priv->cursorlist, 
+							     g_strdup (sv[0]),
+							     pixbuf);
+					g_free (cname);
+					g_strfreev (sv);
+				}
+				g_free (path);
 			}
-		}
-		g_free (xpm_data_file);
+		} 
+		g_free (cursor_dirname);
+		if (cursor_dir) g_dir_close (cursor_dir);
 	}
-	
+	/* don't fallover to source cursor here, we haven't initialized X yet */
+	magnifier_set_cursor_pixmap_by_name (magnifier, "default", FALSE);
 	magnifier_transform_cursor (magnifier);
 }
 
@@ -409,28 +767,12 @@ magnifier_get_property (BonoboPropertyBag *bag,
 	
 	switch (arg_id) {
 	case MAGNIFIER_SOURCE_SIZE_PROP:
-		rect_bounds.x1 = 0;
-		rect_bounds.x2 = gdk_screen_get_width (
-			gdk_display_get_screen (magnifier->source_display,
-						magnifier->source_screen_num));
-		rect_bounds.y1 = 0;
-		rect_bounds.y2 = gdk_screen_get_height (
-			gdk_display_get_screen (magnifier->source_display,
-						magnifier->source_screen_num));
-		BONOBO_ARG_SET_GENERAL (arg, rect_bounds,
+		BONOBO_ARG_SET_GENERAL (arg, magnifier->source_bounds,
 					TC_GNOME_Magnifier_RectBounds,
 					GNOME_Magnifier_RectBounds, NULL);
 		break;
 	case MAGNIFIER_TARGET_SIZE_PROP:
-		rect_bounds.x1 = 0;
-		rect_bounds.x2 = gdk_screen_get_width (
-			gdk_display_get_screen (magnifier->target_display,
-						magnifier->target_screen_num));
-		rect_bounds.y1 = 0;
-		rect_bounds.y2 = gdk_screen_get_height (
-			gdk_display_get_screen (magnifier->target_display,
-						magnifier->target_screen_num));
-		BONOBO_ARG_SET_GENERAL (arg, rect_bounds,
+	    	BONOBO_ARG_SET_GENERAL (arg, magnifier->target_bounds,
 					TC_GNOME_Magnifier_RectBounds,
 					GNOME_Magnifier_RectBounds, NULL);
 
@@ -439,7 +781,8 @@ magnifier_get_property (BonoboPropertyBag *bag,
 		BONOBO_ARG_SET_STRING (arg, magnifier->cursor_set);
 		break;
 	case MAGNIFIER_CURSOR_SIZE_PROP:
-		BONOBO_ARG_SET_INT (arg, magnifier->cursor_size);
+		BONOBO_ARG_SET_INT (arg, magnifier->cursor_size_x);
+		BONOBO_ARG_SET_INT (arg, magnifier->cursor_size_y);
 		break;
 	case MAGNIFIER_CURSOR_ZOOM_PROP:
 		BONOBO_ARG_SET_FLOAT (arg, magnifier->cursor_scale_factor);
@@ -483,6 +826,7 @@ magnifier_set_property (BonoboPropertyBag *bag,
 			gpointer user_data)
 {
 	Magnifier *magnifier = user_data;
+	GNOME_Magnifier_RectBounds rect_bounds;
 	gchar *full_display_string;
 	
 	switch (arg_id) {
@@ -499,6 +843,7 @@ magnifier_set_property (BonoboPropertyBag *bag,
 				gdk_display_get_screen (
 					magnifier->source_display,
 					magnifier->source_screen_num));
+		magnifier_get_display_rect_bounds (magnifier, &magnifier->source_bounds, FALSE);
 		break;
 	case MAGNIFIER_TARGET_DISPLAY_PROP:
 		full_display_string = BONOBO_ARG_GET_STRING (arg);
@@ -513,9 +858,15 @@ magnifier_set_property (BonoboPropertyBag *bag,
 					       gdk_display_get_screen (
 						       magnifier->target_display,
 						       magnifier->target_screen_num));
+		magnifier_get_display_rect_bounds (magnifier, &magnifier->target_bounds, TRUE);
+		magnifier_get_display_rect_bounds (magnifier, &magnifier->source_bounds, FALSE);
+		magnifier_reparent_zoom_regions (magnifier);
 		break;
 	case MAGNIFIER_SOURCE_SIZE_PROP:
-		bonobo_exception_set (ev, ex_Bonobo_PropertyBag_ReadOnly);
+	        magnifier->source_bounds = BONOBO_ARG_GET_GENERAL (arg,
+								   TC_GNOME_Magnifier_RectBounds,
+								   GNOME_Magnifier_RectBounds,
+								   NULL);
 		break;
 	case MAGNIFIER_TARGET_SIZE_PROP:
 	        magnifier->target_bounds = BONOBO_ARG_GET_GENERAL (arg,
@@ -532,11 +883,11 @@ magnifier_set_property (BonoboPropertyBag *bag,
 		
 		break;
 	case MAGNIFIER_CURSOR_SET_PROP:
-		magnifier->cursor_set = g_strdup (BONOBO_ARG_GET_STRING (arg));
-		magnifier_init_cursor (magnifier);
+		magnifier_init_cursor_set (magnifier, g_strdup (BONOBO_ARG_GET_STRING (arg)));
 		break;
 	case MAGNIFIER_CURSOR_SIZE_PROP:
-		magnifier->cursor_size = BONOBO_ARG_GET_INT (arg);
+		magnifier->cursor_size_x = BONOBO_ARG_GET_INT (arg);
+		magnifier->cursor_size_y = BONOBO_ARG_GET_INT (arg);
 		magnifier_transform_cursor (magnifier);
 		break;
 	case MAGNIFIER_CURSOR_ZOOM_PROP:
@@ -548,7 +899,7 @@ magnifier_set_property (BonoboPropertyBag *bag,
 								  TC_CORBA_unsigned_long, 
 								  CORBA_unsigned_long, 
 								  NULL);
-		magnifier_init_cursor (magnifier);
+		magnifier_transform_cursor (magnifier);
 		break;
 	case MAGNIFIER_CURSOR_HOTSPOT_PROP:
 		magnifier->cursor_hotspot = BONOBO_ARG_GET_GENERAL (arg,
@@ -614,8 +965,8 @@ impl_magnifier_set_source_display (PortableServer_Servant servant,
 				ev,
 				magnifier);
 
-	/* attach listener to root window of source display */
-	magnifier_set_expose_listener (magnifier, magnifier_get_root (magnifier));
+	/* attach listeners for DAMAGE, "dirty region", XFIXES cursor changes */
+	magnifier_set_extension_listeners (magnifier, magnifier_get_root (magnifier));
 
 	bonobo_arg_release (arg);
 }
@@ -703,9 +1054,7 @@ impl_magnifier_add_zoom_region (PortableServer_Servant servant,
 
 	if (magnifier->zoom_regions == NULL) 
 	{
-		/* first zoom region */
-		/* add an expose handler to root window; for XSun/XVirtualScreen 'dirty region' support */
-		magnifier_set_expose_listener (magnifier, magnifier_get_root (magnifier));
+		magnifier_set_extension_listeners (magnifier, magnifier_get_root (magnifier));
 	}
 
 	/* FIXME: this needs proper lifecycle management */
@@ -793,7 +1142,9 @@ static void
 magnifier_properties_init (Magnifier *magnifier)
 {
 	BonoboArg *def;
-	
+	GNOME_Magnifier_RectBounds rect_bounds;
+	gchar *display_env;
+
 	magnifier->property_bag =
 		bonobo_property_bag_new_closure (
 			g_cclosure_new_object (
@@ -808,13 +1159,14 @@ magnifier_properties_init (Magnifier *magnifier)
 				     BONOBO_OBJECT (magnifier->property_bag));
 
 	def = bonobo_arg_new (BONOBO_ARG_STRING);
-	BONOBO_ARG_SET_STRING (def, getenv ("DISPLAY"));
+	display_env = getenv ("DISPLAY");
+	BONOBO_ARG_SET_STRING (def, display_env);
 
 	bonobo_property_bag_add (magnifier->property_bag,
 				 "source-display-screen",
 				 MAGNIFIER_SOURCE_DISPLAY_PROP,
 				 BONOBO_ARG_STRING,
-				 NULL,
+				 def,
 				 "source display screen",
 				 Bonobo_PROPERTY_WRITEABLE);
 
@@ -822,28 +1174,40 @@ magnifier_properties_init (Magnifier *magnifier)
 				 "target-display-screen",
 				 MAGNIFIER_TARGET_DISPLAY_PROP,
 				 BONOBO_ARG_STRING,
-				 NULL,
+				 def,
 				 "target display screen",
 				 Bonobo_PROPERTY_WRITEABLE);
 
 	bonobo_arg_release (def);
 
+	magnifier_init_display (magnifier, display_env, TRUE);
+	magnifier_init_display (magnifier, display_env, FALSE);
+
+	magnifier_get_display_rect_bounds (magnifier, &rect_bounds, FALSE);
+	def = bonobo_arg_new_from (TC_GNOME_Magnifier_RectBounds, &rect_bounds);
+        
 	bonobo_property_bag_add (magnifier->property_bag,
-				 "source-display-size",
+				 "source-display-bounds",
 				 MAGNIFIER_SOURCE_SIZE_PROP,
 				 TC_GNOME_Magnifier_RectBounds,
-				 NULL,
+				 def,
 				 "source display bounds/size",
-				 Bonobo_PROPERTY_READABLE);
+				 Bonobo_PROPERTY_READABLE |
+				 Bonobo_PROPERTY_WRITEABLE);
+	bonobo_arg_release (def);
 	
+	magnifier_get_display_rect_bounds (magnifier, &rect_bounds, TRUE);
+	def = bonobo_arg_new_from (TC_GNOME_Magnifier_RectBounds, &rect_bounds);
+
 	bonobo_property_bag_add (magnifier->property_bag,
 				 "target-display-bounds",
 				 MAGNIFIER_TARGET_SIZE_PROP,
 				 TC_GNOME_Magnifier_RectBounds,
-				 NULL,
+				 def,
 				 "target display bounds/size",
 				 Bonobo_PROPERTY_READABLE |
 				 Bonobo_PROPERTY_WRITEABLE);
+	bonobo_arg_release (def);
 
 	bonobo_property_bag_add (magnifier->property_bag,
 				 "cursor-set",
@@ -937,7 +1301,11 @@ magnifier_init_window (Magnifier *magnifier)
 		g_object_connect (gtk_widget_new (gtk_window_get_type (),
 						  "user_data", NULL,
 						  "can_focus", FALSE,
+#ifdef MAG_WINDOW_OVERRIDE_REDIRECT
 						  "type", GTK_WINDOW_POPUP,
+#else
+						  "type", GTK_WINDOW_TOPLEVEL,
+#endif
 						  "title", "magnifier",
 						  "allow_grow", TRUE,
 						  "allow_shrink", TRUE,
@@ -960,7 +1328,8 @@ magnifier_init (Magnifier *magnifier)
 	magnifier->zoom_regions = NULL;
 	magnifier->source_screen_num = 0;
 	magnifier->target_screen_num = 0;
-	magnifier->cursor_size = 0;
+	magnifier->cursor_size_x = 0;
+	magnifier->cursor_size_y = 0;
 	magnifier->cursor_scale_factor = 1.0F;
 	magnifier->cursor_color = 0xFF000000;
 	magnifier->crosswire_size = 1;
@@ -970,17 +1339,31 @@ magnifier_init (Magnifier *magnifier)
 	magnifier->cursor_hotspot.y = 0;
 	magnifier->priv->cursor = NULL;
 	magnifier->priv->w = NULL;
+	magnifier->priv->use_source_cursor = TRUE;
+	magnifier->priv->cursorlist = NULL;
 	magnifier_init_window (magnifier);
-	
-	magnifier->cursor_set = "default";
+	magnifier_init_cursor_set (magnifier, "default");
 }
 
 GdkDrawable *
 magnifier_get_cursor (Magnifier *magnifier)
 {
-	if (!magnifier->priv->cursor)
-		magnifier_init_cursor (magnifier);
-
+        if (magnifier->priv->cursor == NULL) {
+	    if ((fixes_event_base == 0) && 
+		strcmp (magnifier->cursor_set, "none")) 
+	    {
+		GdkPixbuf *pixbuf;
+		gchar *default_cursor_filename = 
+		    g_strconcat (CURSORSDIR, "/", "default-cursor.xpm", NULL);
+		pixbuf = gdk_pixbuf_new_from_file (default_cursor_filename, NULL);
+		if (pixbuf) 
+		{
+		    magnifier_set_cursor_from_pixbuf (magnifier, pixbuf);
+		    g_object_unref (pixbuf);
+		}
+		g_free (default_cursor_filename);
+	    }
+        }
 	return magnifier->priv->cursor;
 }
 
